@@ -1,7 +1,8 @@
 /*
   Copyright (C) 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003,
   2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014,
-  2015 Free Software Foundation, Inc.
+  2015, 2016, 2017, 2018, 2019, 2020, 2021 Free Software Foundation,
+  Inc.
 
   This file is part of GNU Inetutils.
 
@@ -217,6 +218,7 @@ domacro (int argc, char *argv[])
 		   * use of '$' was detected.  Present input is the
 		   * dollar sign.
 		   */
+		  /* FALLTHROUGH */
 		default:
 		  *cp2++ = *cp1;	/* Copy present character.  */
 		  break;
