@@ -1,6 +1,7 @@
 /*
   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
-  2010, 2011, 2012 Free Software Foundation, Inc.
+  2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021
+  Free Software Foundation, Inc.
 
   This file is part of GNU Inetutils.
 
@@ -51,6 +52,8 @@ struct system_ifconfig
   {"hwaddr", system_fh_hwaddr}, \
   {"hwtype?", system_fh_hwtype_query}, \
   {"hwtype", system_fh_hwtype}, \
+  {"metric?", system_fh_metric_query}, \
+  {"metric", system_fh_metric}, \
   {"txqlen?", system_fh_txqlen_query}, \
   {"txqlen", system_fh_txqlen}, \
   {"ifstat?", system_fh_ifstat_query}, \
@@ -82,6 +85,8 @@ void system_fh_hwaddr_query (format_data_t form, int argc, char *argv[]);
 void system_fh_hwaddr (format_data_t form, int argc, char *argv[]);
 void system_fh_hwtype_query (format_data_t form, int argc, char *argv[]);
 void system_fh_hwtype (format_data_t form, int argc, char *argv[]);
+void system_fh_metric_query (format_data_t form, int argc, char *argv[]);
+void system_fh_metric (format_data_t form, int argc, char *argv[]);
 void system_fh_txqlen_query (format_data_t form, int argc, char *argv[]);
 void system_fh_txqlen (format_data_t form, int argc, char *argv[]);
 
