@@ -1,6 +1,7 @@
 /*
   Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011,
-  2012 Free Software Foundation, Inc.
+  2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021 Free
+  Software Foundation, Inc.
 
   This file is part of GNU Inetutils.
 
@@ -32,6 +33,7 @@
 # include <ctype.h>
 # include <syslog.h>
 # include <string.h>
+# include <unused-parameter.h>
 
 # include "auth.h"
 # include "misc.h"
@@ -40,6 +42,8 @@
 #  include "encrypt.h"
 # endif
 
+char *dest_realm = NULL;
+
 Shishi_key *enckey = NULL;
 
 static unsigned char str_data[2048] = { IAC, SB, TELOPT_AUTHENTICATION, 0,
@@ -56,14 +60,18 @@ Shishi_ap *auth_handle;
 
 # define DEBUG(c) if (auth_debug_mode) printf c
 
+/* Callback from consumer.  */
+extern void printsub (char, unsigned char *, int);
+
 static int
-Data (TN_Authenticator * ap, int type, unsigned char *d, int c)
+Data (TN_Authenticator * ap, int type, void * d, int c)
 {
   unsigned char *p = str_data + 4;
   unsigned char *cd = (unsigned char *) d;
 
+  /* Submitted as test data.  */
   if (c == -1)
-    c = strlen (cd);
+    c = strlen ((char *) cd);
 
   if (auth_debug_mode)
     {
@@ -94,7 +102,8 @@ Shishi *shishi_telnet = NULL;
 
 /* FIXME: Reverse return code! */
 int
-krb5shishi_init (TN_Authenticator * ap, int server)
+krb5shishi_init (TN_Authenticator * ap _GL_UNUSED_PARAMETER,
+		 int server)
 {
   if (server)
     str_data[3] = TELQUAL_REPLY;
@@ -128,7 +137,7 @@ delayed_shishi_init (void)
 }
 
 void
-krb5shishi_cleanup (TN_Authenticator * ap)
+krb5shishi_cleanup (TN_Authenticator * ap _GL_UNUSED_PARAMETER)
 {
   if (shishi_handle == NULL)
     return;
@@ -144,12 +153,17 @@ krb5shishi_send (TN_Authenticator * ap)
   char type_check[2];
   Shishi_tkt *tkt;
   Shishi_tkts_hint hint;
-  Shishi_key *subkey;
   int rc;
   char *tmp;
   char *apreq;
   size_t apreq_len;
 
+  if (!UserNameRequested)
+    {
+      DEBUG (("telnet: Kerberos V5: no user name supplied\r\n"));
+      return 0;
+    }
+
   if (!delayed_shishi_init ())
     {
       DEBUG (("telnet: Kerberos V5: shishi initialization failed\r\n"));
@@ -157,9 +171,38 @@ krb5shishi_send (TN_Authenticator * ap)
     }
 
   tmp = malloc (strlen ("host/") + strlen (RemoteHostName) + 1);
-  sprintf (tmp, "host/%s", RemoteHostName);
+  if (tmp == NULL)
+    {
+      DEBUG (("telnet: Kerberos V5: shishi memory allocation failed\r\n"));
+      return 0;
+    }
+
+  /* Check for Kerberos prefix in principal name.  */
+  if (strchr (RemoteHostName, '/'))
+    strcpy (tmp, RemoteHostName);
+  else
+    sprintf (tmp, "host/%s", RemoteHostName);
+
   memset (&hint, 0, sizeof (hint));
   hint.server = tmp;
+  hint.client = UserNameRequested;
+
+  if (dest_realm && *dest_realm)
+    shishi_realm_default_set (shishi_handle, dest_realm);
+  else
+    {
+      /* Retrieve realm assigned to this server as per configuration.  */
+      char *p = strchr (RemoteHostName, '/');
+
+      if (p)
+	++p;
+      else
+	p = RemoteHostName;
+
+      shishi_realm_default_set (shishi_handle,
+				shishi_realm_for_server (shishi_handle, p));
+    }
+
   tkt = shishi_tkts_get (shishi_tkts_default (shishi_handle), &hint);
   free (tmp);
   if (!tkt)
@@ -171,12 +214,6 @@ krb5shishi_send (TN_Authenticator * ap)
   if (auth_debug_mode)
     shishi_tkt_pretty_print (tkt, stdout);
 
-  if (!UserNameRequested)
-    {
-      DEBUG (("telnet: Kerberos V5: no user name supplied\r\n"));
-      return 0;
-    }
-
   if ((ap->way & AUTH_HOW_MASK) == AUTH_HOW_MUTUAL)
     ap_opts = SHISHI_APOPTIONS_MUTUAL_REQUIRED;
   else
@@ -259,18 +296,18 @@ krb5shishi_send (TN_Authenticator * ap)
 }
 
 # ifdef ENCRYPTION
-void
+static void
 shishi_init_key (Session_Key * skey, int type)
 {
   int32_t etype = shishi_key_type (enckey);
 
-  if (etype == SHISHI_DES_CBC_CRC ||
-      etype == SHISHI_DES_CBC_MD4 || etype == SHISHI_DES_CBC_MD5)
+  if (etype == SHISHI_DES_CBC_CRC || etype == SHISHI_DES_CBC_MD4
+      || etype == SHISHI_DES_CBC_MD5)
     skey->type = SK_DES;
   else
     skey->type = SK_OTHER;
   skey->length = shishi_key_length (enckey);
-  skey->data = shishi_key_value (enckey);
+  skey->data = (unsigned char *) shishi_key_value (enckey);
 
   encrypt_session_key (skey, type);
 }
@@ -291,7 +328,7 @@ krb5shishi_reply (TN_Authenticator * ap, unsigned char *data, int cnt)
     {
     case KRB_REJECT:
       if (cnt > 0)
-	printf ("[ Kerberos V5 refuses authentication because %.*s ]\r\n",
+	printf ("[ Kerberos V5 rejects authentication: %.*s ]\r\n",
 		cnt, data);
       else
 	printf ("[ Kerberos V5 refuses authentication ]\r\n");
@@ -319,13 +356,18 @@ krb5shishi_reply (TN_Authenticator * ap, unsigned char *data, int cnt)
 		" (server authenticated)" : " (server NOT authenticated)");
       else
 	printf ("[ Kerberos V5 accepts you ]\r\n");
+
       auth_finished (ap, AUTH_USER);
+      /* This was last access to handle on behalf of the client.  */
+      shishi_done (shishi_handle);
+      shishi_handle = NULL;
       break;
 
     case KRB_RESPONSE:
       if ((ap->way & AUTH_HOW_MASK) == AUTH_HOW_MUTUAL)
 	{
-	  if (shishi_ap_rep_verify_der (auth_handle, data, cnt) != SHISHI_OK)
+	  if (shishi_ap_rep_verify_der (auth_handle, (char *) data, cnt)
+	      != SHISHI_OK)
 	    {
 	      printf ("[ Mutual authentication failed ]\r\n");
 	      auth_send_retry ();
@@ -354,44 +396,37 @@ krb5shishi_reply (TN_Authenticator * ap, unsigned char *data, int cnt)
 }
 
 int
-krb5shishi_status (TN_Authenticator * ap, char *name, int level)
+krb5shishi_status (TN_Authenticator * ap _GL_UNUSED_PARAMETER,
+		   char *name, size_t len, int level)
 {
-  char *cname;
-  int cnamelen;
-  int rc;
   int status;
 
   if (level < AUTH_USER)
     return level;
 
-  rc = shishi_encticketpart_client
-    (shishi_handle,
-     shishi_tkt_encticketpart (shishi_ap_tkt (auth_handle)),
-     &cname, &cnamelen);
-
   if (UserNameRequested
-      && rc == SHISHI_OK
-      && cnamelen == strlen (UserNameRequested)
-      && memcmp (UserNameRequested, cname, cnamelen) == 0)
+      && shishi_authorized_p (shishi_handle,
+			      shishi_ap_tkt (auth_handle),
+			      UserNameRequested))
     {
       /* FIXME: Check buffer length */
-      strcpy (name, UserNameRequested);
+      strncpy (name, UserNameRequested, len);
       status = AUTH_VALID;
     }
   else
     status = AUTH_USER;
-  free (cname);
+
   return status;
 }
 
-int
+static int
 krb5shishi_is_auth (TN_Authenticator * a, unsigned char *data, int cnt,
 		    char *errbuf, int errbuflen)
 {
-  Shishi_key *key, *key2;
+  Shishi_key *key;
   int rc;
-  char *cnamerealm;
-  int cnamerealmlen;
+  char *cnamerealm, *server = NULL, *realm = NULL;
+  size_t cnamerealmlen;
 # ifdef ENCRYPTION
   Session_Key skey;
 # endif
@@ -402,6 +437,93 @@ krb5shishi_is_auth (TN_Authenticator * a, unsigned char *data, int cnt,
       return 0;
     }
 
+  /* Enable use of `~/.k5login'.  */
+  if (shishi_check_version ("1.0.2"))	/* Faulty in version 1.0.1.  */
+    {
+      rc = shishi_cfg_authorizationtype_set (shishi_handle, "k5login basic");
+      if (rc != SHISHI_OK)
+	{
+	  snprintf (errbuf, errbuflen,
+		    "Cannot initiate authorization types: %s",
+		    shishi_error (shishi_handle));
+	  return rc;
+	}
+    }
+
+  if (ServerPrincipal && *ServerPrincipal)
+    {
+      rc = shishi_parse_name (shishi_handle, ServerPrincipal,
+			      &server, &realm);
+      if (rc != SHISHI_OK)
+	{
+	  snprintf (errbuf, errbuflen,
+		    "Cannot parse server principal name: %s",
+		    shishi_strerror (rc));
+	  return 1;
+	}
+      if (realm)
+	shishi_realm_default_set (shishi_handle, realm);
+
+      /* Reclaim an empty server part.  */
+      if (server && !*server)
+	{
+	  free (server);
+	  server = NULL;
+	}
+    }
+
+  if (!server)
+    {
+      server = malloc (strlen ("host/") + strlen (LocalHostName) + 1);
+      if (server)
+	sprintf (server, "host/%s", LocalHostName);
+    }
+
+  if (server)
+    {
+      /* Two possible action on `server':
+       *   "srv.local"    :  rewrite as "host/srv.local"
+       *   "tn/srv.local" :  accept as is
+       */
+      char *p = strchr (server, '/');
+
+      if (!p)
+	{
+	  p = server;
+	  server = malloc (strlen ("host/") + strlen (p) + 1);
+	  if (!server)
+	    {
+	      free (p);		/* This old `server'.  */
+	      snprintf (errbuf, errbuflen,
+			"Cannot allocate memory for server name");
+	      return 1;
+	    }
+	  sprintf (server, "host/%s", p);
+	}
+
+      if (realm)
+	key = shishi_hostkeys_for_serverrealm (shishi_handle,
+					       server, realm);
+      else
+	/* Enforce a search with the known default realm.  */
+	key = shishi_hostkeys_for_serverrealm (shishi_handle,
+			server, shishi_realm_default (shishi_handle));
+
+      free (server);
+    }
+  else
+    key = shishi_hostkeys_for_localservicerealm (shishi_handle,
+						 "host", realm);
+
+  free (realm);
+
+  if (key == NULL)
+    {
+      snprintf (errbuf, errbuflen, "Could not find key: %s",
+		shishi_error (shishi_handle));
+      return 1;
+    }
+
   rc = shishi_ap (shishi_handle, &auth_handle);
   if (rc != SHISHI_OK)
     {
@@ -411,7 +533,7 @@ krb5shishi_is_auth (TN_Authenticator * a, unsigned char *data, int cnt,
       return 1;
     }
 
-  rc = shishi_ap_req_der_set (auth_handle, data, cnt);
+  rc = shishi_ap_req_der_set (auth_handle, (char *) data, cnt);
   if (rc != SHISHI_OK)
     {
       snprintf (errbuf, errbuflen,
@@ -420,18 +542,10 @@ krb5shishi_is_auth (TN_Authenticator * a, unsigned char *data, int cnt,
       return 1;
     }
 
-  key = shishi_hostkeys_for_localservice (shishi_handle, "host");
-  if (key == NULL)
-    {
-      snprintf (errbuf, errbuflen, "Could not find key:\n%s\n",
-		shishi_error (shishi_handle));
-      return 1;
-    }
-
   rc = shishi_ap_req_process (auth_handle, key);
   if (rc != SHISHI_OK)
     {
-      snprintf (errbuf, errbuflen, "Could not process AP-REQ: %s\n",
+      snprintf (errbuf, errbuflen, "Could not process AP-REQ: %s",
 		shishi_strerror (rc));
       return 1;
     }
@@ -445,7 +559,7 @@ krb5shishi_is_auth (TN_Authenticator * a, unsigned char *data, int cnt,
       rc = shishi_ap_rep_der (auth_handle, &der, &derlen);
       if (rc != SHISHI_OK)
 	{
-	  snprintf (errbuf, errbuflen, "Error DER encoding aprep: %s\n",
+	  snprintf (errbuf, errbuflen, "Error DER encoding aprep: %s",
 		    shishi_strerror (rc));
 	  return 1;
 	}
@@ -454,13 +568,13 @@ krb5shishi_is_auth (TN_Authenticator * a, unsigned char *data, int cnt,
       free (der);
     }
 
-  rc = shishi_encticketpart_crealm (shishi_handle,
-				    shishi_tkt_encticketpart (shishi_ap_tkt
-							      (auth_handle)),
-				    &cnamerealm, &cnamerealmlen);
+  rc = shishi_encticketpart_clientrealm (
+		shishi_handle,
+		shishi_tkt_encticketpart (shishi_ap_tkt (auth_handle)),
+		&cnamerealm, &cnamerealmlen);
   if (rc != SHISHI_OK)
     {
-      snprintf (errbuf, errbuflen, "Error getting authenticator name: %s\n",
+      snprintf (errbuf, errbuflen, "Error getting authenticator name: %s",
 		shishi_strerror (rc));
       return 1;
     }
@@ -470,6 +584,11 @@ krb5shishi_is_auth (TN_Authenticator * a, unsigned char *data, int cnt,
   free (cnamerealm);
   auth_finished (a, AUTH_USER);
 
+  /* Make sure that shishi_handle is still valid,
+   * it must not be released in auth_finish()!
+   * The server side will make reference to it
+   * later on.  */
+
 # ifdef ENCRYPTION
   if (enckey)
     {
@@ -517,8 +636,6 @@ krb5shishi_is (TN_Authenticator * ap, unsigned char *data, int cnt)
   int r = 0;
   char errbuf[512];
 
-  puts ("krb5shishi_is");
-
   if (cnt-- < 1)
     return;
   errbuf[0] = 0;
@@ -569,13 +686,11 @@ req_type_str (int type)
 
 void
 krb5shishi_printsub (unsigned char *data, int cnt,
-		     unsigned char *buf, int buflen)
+		     char *buf, int buflen)
 {
   char *p;
   int i;
 
-  puts ("krb5shishi_printsub");
-
   buf[buflen - 1] = '\0';	/* make sure its NULL terminated */
   buflen -= 1;
 
